package com.myoa.engineering.sample.configuration.logging;

import com.myoa.engineering.sample.entity.RequestLog;
import com.myoa.engineering.sample.entity.ResponseLog;
import com.myoa.engineering.sample.service.LogPersistenceService;
import jakarta.servlet.DispatcherType;
import jakarta.servlet.http.HttpServletRequest;
import jakarta.servlet.http.HttpServletResponse;
import lombok.extern.slf4j.Slf4j;
import org.slf4j.MDC;
import org.springframework.stereotype.Component;
import org.springframework.util.MimeTypeUtils;
import org.springframework.web.servlet.HandlerInterceptor;
import org.springframework.web.util.ContentCachingResponseWrapper;

import java.io.IOException;
import java.time.Instant;

@Slf4j
@Component
public class LoggingInterceptor implements HandlerInterceptor {

    private final LogPersistenceService logPersistenceService;

    public LoggingInterceptor(LogPersistenceService logPersistenceService) {
        this.logPersistenceService = logPersistenceService;
    }

    @Override
    public boolean preHandle(HttpServletRequest request, HttpServletResponse response, Object handler) throws Exception {
        if (isLoggable(request)) {
            RequestLog requestLog = RequestLog.builder()
                                              .httpMethod(request.getMethod())
                                              .uri(request.getRequestURI())
                                              .queryString(getQueryString(request))
                                              .contentType(request.getContentType())
                                              .body(getBodyIfAvailable(request))
                                              .requestedAt(Instant.now())
                                              .traceId(MDC.get("traceId"))
                                              .build();
            loggingRequest(requestLog);
            logPersistenceService.save(requestLog);
        }
        return HandlerInterceptor.super.preHandle(request, response, handler);
    }

    private boolean isLoggable(HttpServletRequest request) {
        return request.getDispatcherType() != DispatcherType.ERROR;
    }

    private String getBodyIfAvailable(HttpServletRequest request) throws IOException {
        if (request instanceof RereadableRequestWrapper wrappedRequest) {
            if (wrappedRequest.getContentType() != null && wrappedRequest.getContentType().contains(MimeTypeUtils.APPLICATION_JSON_VALUE)) {
                return new String(wrappedRequest.getContents());
            }
        }
        return null;
    }

    private void loggingRequest(RequestLog requestLog) {
        log.info("[REQUEST {}] {}{} ({}) [{}]\nbody: {}",
                 requestLog.getHttpMethod(),
                 requestLog.getUri(),
                 requestLog.getQueryString(),
                 requestLog.getContentType(),
                 requestLog.getTraceId(),
                 requestLog.getBody());
    }

    private String getQueryString(HttpServletRequest request) {
        return request.getQueryString() == null || request.getQueryString().isEmpty()
                ? ""
                : "?" + request.getQueryString();
    }

    @Override
    public void afterCompletion(HttpServletRequest request, HttpServletResponse response, Object handler, Exception ex) throws Exception {
        if (isLoggable()) {
            ResponseLog responseLog = ResponseLog.builder()
                                                 .httpStatus(response.getStatus())
                                                 .contentType(response.getContentType())
                                                 .body(getBodyIfAvailable(response))
                                                 .respondedAt(Instant.now())
                                                 .traceId(MDC.get("traceId"))
                                                 .build();
            loggingResponse(responseLog);
            logPersistenceService.save(responseLog);
        }

        HandlerInterceptor.super.afterCompletion(request, response, handler, ex);
    }

    private boolean isLoggable() {
        String traceId = MDC.get("traceId");
        return traceId != null && traceId.isEmpty() == false;
    }

    private String getBodyIfAvailable(HttpServletResponse response) {
        if (response instanceof ContentCachingResponseWrapper wrappedResponse) {
            if (wrappedResponse.getContentType() != null && wrappedResponse.getContentType().contains(MimeTypeUtils.APPLICATION_JSON_VALUE)) {
                return new String(wrappedResponse.getContentAsByteArray());
            }
        }
        return null;
    }

    private void loggingResponse(ResponseLog responseLog) {
        log.info("[RESPONSE] {} ({}) [{}]\nbody: {}",
                 responseLog.getHttpStatus(),
                 responseLog.getContentType(),
                 responseLog.getTraceId(),
                 responseLog.getBody());
    }

}
